try:

    from IPython.display import display
    import ipywidgets as widgets
    from ipywidgets import HBox, Label, VBox
    from ipywidgets import Button, Layout, HTML
    from traitlets import MetaHasTraits

except ModuleNotFoundError:

    class VBox():
        pass

    class MetaHasTraits():
        pass
        

class ParamChoiceWidget(VBox):
    
    def __init__(self, name, options, default_value, optional=False, widget=None):

        if not isinstance(VBox, MetaHasTraits):
            return

        self.name = name
        
        if optional:
            label = widgets.HTML(f"<b><font size=2 color='grey'>{self.name}</b>")
        else:
            label = widgets.HTML(f"<b><font size=2 color='black'>{self.name}</b>")

        if widget:
            self.widget = widget
        else:
            self.widget = widgets.RadioButtons(options=options, disabled=False)

        self.widget.value = default_value

        children = [label, self.widget]
        super().__init__(children=children)
        
    def get_value(self):
        
        return self.widget.value

class ParamSetterWidget(VBox):

    def __init__(self, name, widget, default_value):

        if not isinstance(VBox, MetaHasTraits):

            return

        self.name = name

        if default_value:
            label = widgets.HTML(f"<b><font size=2 color='blue'>{self.name}</b>")
        else:
            label = widgets.HTML(f"<b><font size=2 color='black'>{self.name}</b>")

        self.widget = widget

        self.widget.value = default_value

        children = [label, self.widget]
        super().__init__(children=children)

        
class Custom1(VBox):
    
    def __init__(self):

        if not isinstance(VBox, MetaHasTraits):

            return

        self.name = 'custom1'
        
        layout = {'width': '600px'}
        style = {'description_width': 'initial'}

        w1=widgets.IntRangeSlider(
            value=[10, 150],
            min=0,
            max=300,
            step=1,
            description='Regs search range:',
            disabled=False,
            continuous_update=False,
            orientation='horizontal',
            readout=True,
            readout_format='d',
            layout=layout,
            style=style
        )

        layout = {'width': '150px'}
        style = {'description_width': 'initial'}
        w2=widgets.Dropdown(
            options=[str(elem) for elem in list(range(1, 10))],
            value='1',
            description='Search gaps:',
            disabled=False,
            layout=layout
        )


        layout = {'width': '600px'}
        style = {'description_width': 'initial'}

        w3=widgets.IntRangeSlider(
            value=[5, 60],
            min=0,
            max=120,
            step=1,
            description='Days search range:',
            disabled=False,
            continuous_update=False,
            orientation='horizontal',
            readout=True,
            readout_format='d',
            layout=layout,
            style=style
        )

        layout = {'width': '150px'}
        style = {'description_width': 'initial'}
        w4=widgets.Dropdown(
            options=[str(elem) for elem in list(range(1, 10))],
            value='1',
            description='Search gaps:',
            disabled=False,
            layout=layout
        )
        
        h1 = HBox([w1, w2])
        h2 = HBox([w3, w4])

        label = widgets.HTML(f"<b><font size=2 color='black'>{'search_space'}</b>")     
        
        children = [label, h1, h2]
        super().__init__(children=children)
        
    def get_value(self):
        
        region_range = self.children[0].children[0].value
        region_jump = self.children[0].children[1].value

        day_range = self.children[1].children[0].value
        day_jump = self.children[1].children[1].value
        
        d = {'num_days': list(range(*day_range, int(day_jump))),
             'num_regions': list(range(*region_range, int(region_jump)))}
        
        return d
       
class ParamTextWidget(VBox):
    
    def __init__(self, name, example=None, optional=False, widget=None):

        if not isinstance(VBox, MetaHasTraits):

            return

        self.name = name
        
        style = style = {'description_width': 'initial'}
        layout = Layout(display='flex', 
                        flex_flow='column', 
                        align_items='flex-start', 
                        border=None, 
                        width='50%',
                        align_content='flex-start')
        
        if optional:
            label = widgets.HTML(f"<b><font size=2 color='grey'>{self.name}</b>")
        else:
            label = widgets.HTML(f"<b><font size=2 color='black'>{self.name}</b>")
        
        if widget:
            self.widget = widget
        else:
            if example is None:
                self.widget = widgets.Text(style=style, layout=layout)
            else:
                self.widget = widgets.Text(description='E.g. {}: '.format(example), style=style, layout=layout)
        
        children = [label, self.widget]
        super().__init__(children=children)
        
    def get_value(self):
        
        return self.widget.value




from collections import defaultdict

from .tree_node import TreeNode

def construct_tree(parsed_node_data):

    parent_nodes = {}

    node_dicts = defaultdict(dict)

    root = TreeNode(-2, 'root', None, 'root')
    parent_nodes[-2] = root

    for node_data in parsed_node_data:

        node_name, primary_type, secondary_type, depth, default_value = node_data

        if primary_type=='topic':
            depth = -1
            current_topic = node_name  # Topic has to be on the top of the dsl

        node = TreeNode(depth, node_name, default_value, primary_type, secondary_type)
        parent_nodes[depth] = node

        if not primary_type=='topic':

            node_dicts[current_topic][node_name] = node

        parent_nodes[depth-1].add_child(node)

    return root, node_dicts


def display_tree(root):
        
    _display_tree(root)
    
def _display_tree(node):

    depth = node.depth
    node_type = node.primary_type
    node_name = node.name

    if node_type != 'root':

        if node_type == 'topic':
            print()

            depth += 1

        print('    '*depth, node_name, ':', node.primary_type, node.secondary_type, node.default_value)

    for child in node.children:

        _display_tree(child)

        


from .tree.tree_parser import parse_dsl
from .tree.tree_builder import construct_tree
from .builder.access_object import AccessObject
from .builder.get_control_panel import get_control_panel

try:
    
    from IPython.display import display

except ModuleNotFoundError:

    pass

import inspect
import os
import functools
import warnings


class ArgtellerClassDecorator():
    
    def __init__(self, dsl, override=False):
        
        if os.path.exists(dsl):
            with open(dsl) as f:
                self.dsl = f.read()

        elif isinstance(dsl, str):
            self.dsl = dsl
            
        self.override = override
        
    def __call__(self, cls):
        
        class Wrapped(cls):
            
            def __init__(cls_self, *args, **kwargs):
                
                source_obj_dict = {}
                
                # The signature of the class being decorated.
                original_signature = inspect.signature(cls.__init__)
                
                params = list(original_signature.parameters.values())
                
                param_names = [param.name for param in params]
                param_types = [param.kind for param in params]
                
                # Because the inner __init__ method signature only consists
                # of VAR_POSITIONAL and VAR_KEYWORD type parameter, we need
                # to check manually.
                
                # If **kwargs is not in the original_signature,
                # we cannot accept kwargs not in the param_names.
                if not inspect.Parameter.VAR_KEYWORD in param_types:
                    
                    for key, value in kwargs.items():
                    
                        if not key in param_names:

                            raise TypeError("__init__() got an unexpected keyword argument '{}'".format(
                                key))
                
                # If *args is not in the original_signagure,
                # we cannot accept args longer than there are
                # POSITIONAL_OR_KEYWORD type args in the signature.
                if not inspect.Parameter.VAR_POSITIONAL in param_types:
                    
                    num_pos_or_kw = len([param_type for param_type in param_types if
                                         param_type==inspect.Parameter.POSITIONAL_OR_KEYWORD])
                    
                    if len(args) > num_pos_or_kw - 1:  # -1 for the implicit self argument
                        
                        raise TypeError("__init__() takes {} positional arguments but {} were given".format(
                            num_pos_or_kw, len(args) + 1))  # +1 to count for the implicit self argument
                
                # Check the user passed arguments at the __init__ method invocation.
                check_pos_args = []
                
                for i, param in enumerate(params):
                    
                    if i==0:  # Skip the implicit self argument.
                        continue
                    
                    if param.kind==inspect.Parameter.POSITIONAL_OR_KEYWORD:
                        
                        if len(args)>=i:
                            
                            setattr(cls_self, param.name, args[i-1])
                            
                        elif param.name in kwargs:
                            
                            setattr(cls_self, param.name, kwargs[param.name])
                            del kwargs[param.name]
                            
                        else:
                            
                            if param.default==inspect._empty:
                            
                                check_pos_args.append(param.name)
                            
                            # The Method decorator will source from the source
                            # object here. But for Class decorator, because the 
                            # widgets are dynamic, we cannot do that.
                            
                            # We will only check to see if the missing argument
                            # is at least found in the widget param namespace.
                        
                # The Method decorator can check the missing_positional_arguments
                # to throw TypeError missing argument exception. But with the 
                # Class decorator, we cannot do that because we are waiting on the
                # user to interact with the widget. 
                # So instead, we will rely on the requirement signals of the widgets.

                parsed_node_data = parse_dsl(self.dsl)
                root, node_dicts = construct_tree(parsed_node_data)
                cls_self.__access_object__ = AccessObject(root, node_dicts)

                if cls_self.__access_object__.module_found:

                    cls_self.__control_panel__ = get_control_panel(cls_self.__access_object__)
                    display(cls_self.__control_panel__)
                
                    widget_params = cls_self.__access_object__.get_params()
                    cls_self.topic = None

                else:

                    warnings.filterwarnings('always')

                    warnings.warn("Please install 'IPython' and 'ipywidgets' modules to enable widgets.", ImportWarning)

                    warnings.filterwarnings(action='ignore', category=DeprecationWarning, module='ipykernel')
                    
                    widget_params = []

                # Below missing positional arguments cannot be found in the widget 
                # param namespace.
                
                missing_pos_args = []
                
                for check_pos_arg in check_pos_args:
                    
                    if check_pos_arg not in widget_params:
                        
                        missing_pos_args.append("'{}'".format(check_pos_arg))
                        
                if len(missing_pos_args) > 0:

                    missing_args = " and ".join(missing_pos_args)

                    raise TypeError("__init__() missing {} required positional arguments: {} !".format(
                        len(missing_pos_args), missing_args))

            def __getattr__(cls_self, elem):
                """This magic method is invoked when the __getattribute__ 
                magic method throws an exception. This is a natural way to
                query the widgets when the user has not supplied a required
                argument, or when the user queries for a parameter that is 
                not specified in the original signature.
                """
                if cls_self.__access_object__.module_found and elem in cls_self.__access_object__.get_params():

                    value = cls_self.__access_object__.get_value(elem, cls_self.topic)

                    if value is None:
                        return value

                    try:    
                        value = int(value)
                        return value
                    except ValueError:
                        pass
                    
                    try:
                        value = float(value)
                        return value
                    except ValueError:
                        pass
                    
                    try:
                        value = eval(value)
                        return value
                    except (SyntaxError, NameError):
                        pass

                    if value=='':
                        return None
                    
                    return value

                else:
                    
                    raise AttributeError("'{}' object has no attribute '{}'!".format(cls.__name__, elem))
                
            def __settopic__(cls_self, topic):
                
                cls_self.topic = topic
                
            def __resettopic__(cls_self):
                
                cls_self.topic = None

            def __getparams__(cls_self):

                return cls_self.__access_object__.get_params(cls_self.topic)

            def __getvalue__(cls_self, param):

                return cls_self.__access_object__.get_value(param, cls_self.topic)

        return Wrapped

def ArgtellerMethodDecorator(source_name, topic=None):
    
    def decorator(func):

        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            
            original_signature = inspect.signature(func)
            
            new_args = [self]
    
            params = list(original_signature.parameters.values())
            
            has_VAR_POSITIONAL = inspect.Parameter.VAR_POSITIONAL in [param.kind for param in params]
            has_VAR_KEYWORD = inspect.Parameter.VAR_KEYWORD in [param.kind for param in params]
        
            
            if source_name in kwargs:
                
                __source_obj__ = kwargs[source_name]
                
                __source_obj__.__settopic__(topic)
                
                del kwargs[source_name]
                
            else:
                
                __source_obj__ = None

            missing_positional_arguments = []
            
            num_hard_pos = 0
            num_pos = 0
                
            for i, param in enumerate(params):      
                
                if i==0:  # Skip the implicit self argument.
                    continue
 
                if param.kind==inspect.Parameter.POSITIONAL_OR_KEYWORD:
                    
                    # Keep track of the number of positional arguments to check
                    # excessive argument input error.
                    if param.default==inspect._empty:
                        num_hard_pos += 1
                    num_pos += 1
                    
                    # We can assume that the first args belong to the
                    # position argument list.
                    if len(args)>=i:
                        
                        new_args.append(args[i-1])
                    
                    elif param.name in kwargs:
                        
                        new_args.append(kwargs[param.name])
                        del kwargs[param.name]
                        
                    else:
                        
                        if __source_obj__ is not None:
                        
                            if param.name in __source_obj__.__getparams__():

                                new_args.append(__source_obj__.__getvalue__(param.name))

                        else:
                            
                            if param.default==inspect._empty:
                                
                                missing_positional_arguments.append("'{}'".format(param.name))
                            
                            new_args.append(param.default)
                            
            if not has_VAR_POSITIONAL and num_pos < len(args):
                
                # Add 1 for the implicit self argument.
                if num_hard_pos > 0:
                    
                    raise TypeError("__init__() takes {} positional arguments but {} were given".format(
                        num_pos+1, len(args)+1))
                    
                else:
                    
                    raise TypeError("__init__() takes from {} to {} positional arguments but {} were given".format(
                        num_hard_pos+1, num_pos+1, len(args)+1))

            if len(missing_positional_arguments) > 0:

                missing_args = " and ".join(missing_positional_arguments)

                raise TypeError("__init__() missing {} required positional arguments: {} !".format(len(missing_positional_arguments), missing_args))
                        
            cr = func(*new_args, **kwargs) # call original function

            return cr

        return wrapper
    
    return decorator




from .tree_node import TreeNode

import re
from collections import defaultdict


def check_type(line):
    if '/' in line:
        return 'param_setter'
    elif line[0]=='-':
        return 'param'
    elif line[0:2]=='==':
        return 'string_sample'
    elif line[0]=='=':
        return 'option'
    elif line[0]=='+':
        return 'optional'
    elif line[0]=='?':
        return 'boolean'
    elif line[0]=='#':
        return 'custom{}'.format(line[1])
    else:
        return 'topic'

def check_default(line):
    
    line = line.replace(" ", "")
    
    if not ':' in line:
        return line, None
    
    try:
        line1, line2 = line.split(':')

    except:

        print(line)
    
    if line2=='':
        return line1, None
    
    if ',' in line2:
    
        line2, line3 = line2.split(',')

        if line3=='int':
            default_value = int(eval(line2))
        elif line3=='float':
            default_value = float(eval(line2))
        else:
            default_value = line2
            
    else:
        
        try:
            
            default_value = float(line2)
            
            if default_value.is_integer():
                default_value = int(default_value)
                
        except ValueError:
            
            default_value = line2
            
    return line1, default_value

def get_depth(line):
    
    tab_depth = line.count('\t')
    space_depth = int((len(line) - len(line.lstrip(' ')))/4)
    return tab_depth + space_depth

def parse_dsl(dsl):

    parsed_node_data = []

    dsl_lines = dsl.splitlines()
    dsl_lines = dsl_lines + ['']

    depths = [0]

    for idx in range(1, len(dsl_lines)):

        primary_type = None

        prev_line = dsl_lines[idx-1]

        prev_depth = get_depth(prev_line)

        prev_line = prev_line.strip()

        if prev_line=='':
            pass
        else:
            primary_type = check_type(prev_line)

        secondary_type = None

        line = dsl_lines[idx]

        depth = get_depth(line)

        line = line.strip()

        if line=='':
            pass
        elif prev_depth < depth:
            # if we don't have this condition, the deeper choice might
            # think it has shallower param induced from it
            secondary_type = check_type(line)

        name = re.sub('^[\s=+-?#0-9]+', '', prev_line)

        name, default_value = check_default(name)
        
        if (primary_type=='param' or primary_type=='optional') and secondary_type is None:
            secondary_type = 'string'

        if name != '':
            parsed_node_data.append([name, primary_type, secondary_type, prev_depth, default_value])
            
    return parsed_node_data

